/******************************************************************************
*
*      FILE: qs_sl.c
*   VERSION: 4.00
*      DATE: May 2009
* COPYRIGHT: QAS Ltd
*
* C sample code for integrating the QAS Pro Single Line search engine (V4.x).
*
* This integration is suitable for environments where it is not practical to
* detect individual keypresses as they are typed by the operator. Instead, it
* relies on the operator entering text and then pressing the <enter> key in
* order to submit a search or navigate the resulting picklists.
*
* There are 3 files that make up the sample code:
*
* - qs_sl.h    : Definitions, function prototypes, etc.
* - qs_sl.c    : Functions defining recommended address capture process
* - qs_sl_io.c : Functions performing user input and output
*
* The structure Search_t (as defined in qs_sl.h) is used to maintain the search
* context and is passed in as a parameter to all functions. It contains the
* handle to the QAS Pro API, and details of where the operator is with regards
* navigating the picklists. This structure can be amended or extended as
* required by the integrator.
*
* Full error handling has been incorporated including the ability to detect and
* restore a broken client/server connection. Note that not all calls to the QAS 
* Pro API are tested for errors as the sample code uses a form of "lazy" checking
* whereby connection errors are picked up when the pick-list of results is
* updated. It is at this point that the search is aborted.
*
* For simplicity, this process does not support multiple (concurrent) searches
* being performed. If this is a requirement, then it will be necessary to alter
* the logic behind the QS_Startup and QS_Shutdown functions so that multiple
* handles can be created and destroyed.
*
******************************************************************************/


#include "qs_sl.h"



/**************
* main()
*
* Program entry point. This function demonstrates the high-level address
* capture process, calling the key modules to initialise the API, get the
* search parameters, perform a search, manipulate the resulting picklists,
* and finally return the selected (formatted) address.
*
* Return:
*   0 = Success
***************/

int main (void)
{
    Search_t    tSearch;                    /* Holds search context for the address capture process */
    char        sDataSet [DATASET_LENGTH];  /* Data set as selected by the operator */
    char        sSearch [SEARCH_LENGTH];    /* Entered search expression */
    int         iError = 0;                 /* Error code */


    /* The calls to QS_Startup and QS_Shutdown can occur either:  */
    /*   a) when the host application starts-up and shuts-down    */
    /*   b) each time an address is captured                      */
    /* Solution A prevents any potential delay occurring prior to */
    /* the capture process. Solution B is more modular in its     */
    /* approach and minimises resource usage.                     */

    if (QS_Startup (&tSearch, sDataSet))
    {
        /* Loop until the user has no more searches to perform. In   */
        /* a typical integration, the user will only be entering a   */
        /* single search at a time (so an IF construct will normally */
        /* be more appropriate).                                     */

        while (QS_GetSearchParameters (&tSearch, sDataSet, sSearch))
        {
            iError = QS_SingleLineSearch (&tSearch, sDataSet, sSearch);

            if (iError == 0
                &&
                QS_NavigatePickLists (&tSearch))
            {
                /* Multiple layouts can be defined and used if the host */
                /* application uses different formats.                  */

                iError = QS_FormatAddress (&tSearch);
            }

            if (iError != 0)
            {
                QS_ReportError (&tSearch, iError, "Search failed due to error");
            }

            /* Always end the search */
            QS_EndSearch (&tSearch);
        }
    }

    
    /* Always shutdown the system */
    QS_Shutdown (&tSearch);


    return (0);
}
        



/**************
* QS_Startup()
*
* Initialise the QAS Pro API. This function performs the API open,
* storing a handle to the API within the search context structure.
* The call also obtains the active data set that has been opened
* by default.
*
* Input:
*   ptSearch : Search context
*
* Return:
*   0 = Error occurred (will have to resort to manual entry of address)
*   1 = QAS Pro API successfully initialised
***************/

int QS_Startup (Search_t *ptSearch, char *psDataSet)
{
    int   iError = 0;                               /* QAS Pro API error code */
    char  sLayout[SEARCH_LENGTH] = "";              /* Layout to set */


    /* Reset structure to a known state */
    ptSearch->iHandle       = -1;
    ptSearch->iLevel        = -1;
    ptSearch->iAddress      = -1;


    /* Initialise the API and obtain a API handle */
    iError = QA_Open ("", "", &ptSearch->iHandle);

    /* Obtain the data set that has been opened by default */
    if (iError == 0)
    {
        iError = QA_GetActiveData (ptSearch->iHandle, psDataSet, DATASET_LENGTH);
    }

    /* Prompt the user if they wish to specify a layout */
    /* If they do not, it will use the "" layout, which uses the layout */
    /* in [QADefault] section in the configuration file, or an internal */
    /* layout if this is missing */
    if (iError == 0)
    {
        QS_GetLayoutParameter (sLayout);

        iError = QA_SetActiveLayout (ptSearch->iHandle, sLayout);
    }


    if (iError != 0)
    {
        QS_ReportError (ptSearch, iError, "Unable to initialise");
    }

    return (iError == 0 ? 1 : 0);
}




/**************
* QS_Shutdown()
*
* Shutdown the QAS Pro API, releasing all resources.
*
* Input:
*   ptSearch : Search context
*
* Return:
*   None.
***************/

void QS_Shutdown (Search_t *ptSearch)
{
    int     iError = 0;             /* QAS Pro error code */


    /* This call will release all resources associated  */
    /* with an API handle. For safety, this should      */
    /* always be called with the handle returned from   */
    /* the corresponding call to QA_Open ().            */

    iError = QA_Close (ptSearch->iHandle);

    if (iError != 0)
    {
        QS_ReportError (ptSearch, iError, "Error closing an instance");
    }

    /* Set the handle to a deterministic value */
    ptSearch->iHandle = -1;
    

    /* The call to QA_Shutdown () should be the very last   */
    /* function called in the integration. This would       */
    /* typically be when the entire system is stopped.      */

    QA_Shutdown ();
}



/**************
* QS_SingleLineSearch()
*
* Execute a single line search. This function will submit the specified search
* expression and then perform an automatic step-in if appropriate to do so.
* Note that any results pending from a previous search will be discarded.
*
* Input:
*   ptSearch   : Search context
*   psDataSet  : Name of data set (e.g. "GBR","USA")
*   psSearch   : Search expression
*
* Return:
*   0       : Success
*   != 0    : Search failed
***************/

int QS_SingleLineSearch (Search_t *ptSearch, char *psDataSet, char *psSearch)
{
    int     iError      = 0;                    /* QAS Pro error code */


    /* Reset search context */
    ptSearch->iLevel    = -1;
    ptSearch->iAddress  = -1;


    /* Select the requested data set */
    iError = QA_SetActiveData (ptSearch->iHandle, psDataSet);

    
    /* Select the search engine */
    if (iError == 0)
    {
        iError = QA_SetEngine (ptSearch->iHandle, qaengine_SINGLELINE);
    }


    /* Perform search using entered string */
    if (iError == 0)
    {
        iError = QA_Search (ptSearch->iHandle, psSearch);
    }


    /* Remember search history and increment picklist level in context */
    if (iError == 0)
    {
        ptSearch->iLevel ++;

        strcpy (ptSearch->asHistory[ptSearch->iLevel], psSearch);
    }

    return (iError);
}



/**************
* QS_EndSearch()
*
* Ends the current search. It is very important to end one
* search before beginning another, otherwise errors will
* be returned from the API. You can safely call QA_EndSearch
* if a search has not been performed.
*
* Input:
*   ptSearch   : Search context
*
* Return:
*   None
***************/

void QS_EndSearch (Search_t *ptSearch)
{
    int     iError  = 0;                /* QAS Pro API error code */

    iError = QA_EndSearch (ptSearch->iHandle);

    if (iError != 0)
    {
        QS_ReportError (ptSearch, iError, "Ending a search");
    }
}




/**************
* QS_UpdatePickList()
*
* Display all picklist items to the user. This function iterates through the
* picklist, extracting and formatting the address, postcode, and score. The
* relevant display function is then called to output the results to the user.
*
* Input:
*   ptSearch    : Search context
*   iAvailable  : Picklist size
*   iPotential  : Potential matches
*
* Return:
*   0       : Successful
*   != 0    : Failure
***************/

int QS_UpdatePickList (Search_t *ptSearch, int iAvailable, int iPotential)
{
    int     iError      = 0;            /* QAS Pro error code */
    int     iIndex      = 0;            /* Used to iterate through all picklist items */
    char    sLine [SEARCH_LENGTH];      /* Picklist item description from API */
    int     iScore      = 0;            /* Score give to picklist item */
    long    lFlags      = 0;            /* Picklist item flags */
    char    sAddress [SEARCH_LENGTH];   /* Formatted address */
    char    sSymbol [SEARCH_LENGTH];    /* Symbol string */
    char    sScore [SEARCH_LENGTH];     /* Score string */
    char   *psPostcode;                 /* Pointer to postcode in result */


    /* Display picklist header */
    QS_StartPickListDisplay (ptSearch, iPotential);


    /* Iterate for every picklist item */
    for (iIndex = 0; iError == 0 && iIndex < iAvailable; iIndex ++)
    {
        /* Get information about the result */
        iError = QA_GetResult (ptSearch->iHandle, iIndex, sLine, sizeof (sLine), &iScore, &lFlags);

        if (iError == 0)
        {
            /* Add '+' if item can be stepped into */
            if (lFlags & qaresult_CANSTEP)
            {
                strcpy (sSymbol, "+ ");
            }
            else
            {
                strcpy (sSymbol, "  ");
            }

            /* Separate postcode */
            psPostcode = strchr (sLine, '\t');
            
            if (psPostcode != NULL)
            {
                *psPostcode = '\0';
                psPostcode ++;
            }
            else
            {
                psPostcode = "";
            }

            
            /* Ensure address line is the desired width */
            if (strlen (sLine) > ADDRESS_RESULT_WIDTH - 3)
            {
                strcpy (sLine + ADDRESS_RESULT_WIDTH - 3, " ..");
            }

            /* Add score if first level */
            if (ptSearch->iLevel == 0)
            {
                /* Note that it only makes sense to display a */
                /* score at the first picklist level          */

                sprintf(sScore, "%3d%%", iScore);
            }
            else
            {
                sScore[0] = '\0';
            }


            /* Construct picklist string */
            sprintf (sAddress,
                     "%-3d: %-*s %-9s  %-4s",
                     iIndex + 1,
                     ADDRESS_RESULT_WIDTH,
                     sLine,
                     psPostcode,
                     sScore);

            /* Display item to user */
            QS_DisplayPickListItem (sAddress);
        }
    }

    return (iError);
}



/**************
* QS_SelectPickListItem()
*
* Attempt to select a picklist item. If the item is invalid or cannot be
* resolved then a warning is displayed to the user. If the item is not a
* leaf-node address then it is stepped into.
*
* Input:
*   ptSearch   : Search context
*   iAvailable : Picklist size
*   iIndex     : Offset of picklist item being selected
*
* Return:
*   0 = Error occurred
*   1 = Picklist item selected
***************/

int QS_SelectPickListItem (Search_t *ptSearch, int iAvailable, int iIndex)
{
    int     iError      = 0;                /* QAS Pro error code */
    long    lFlags      = 0;                /* Attributes of match */

    char    sHistory [256];                 /* New history text */

    
    /* Check that the chosen picklist is within the boundaries of the picklist */
    if (iIndex < 0 || iIndex >= iAvailable)
    {
        QS_DisplayMessage ("Invalid offset for selected item");

        return 0;
    }


    /* Get details of picklist entry */
    iError = QA_GetResult (ptSearch->iHandle, iIndex, NULL, 0, NULL, &lFlags);

    
    /* Use the partial address of the item as the history */
    if (iError == 0)
    {
        iError = QA_GetResultDetail (ptSearch->iHandle,
                                     iIndex,
                                     qaresultstr_PARTIALADDRESS,
                                     NULL,
                                     sHistory,
                                     sizeof (sHistory));
    }


    /* Decide what to do with selected result */
    if (iError == 0)
    {
        /* Determine whether we should              */
        /* 1. Step in to result                     */
        /* 2. Mark this address as the one to be    */
        /*    formatted and returned.               */
        /* 3. Inform user that they've selected an  */
        /*    informational prompt.                 */
        /* 4. Inform user that they're selected an  */
        /*    unexpandable range. They have to type */
        /*    in the desired value.                 */

        if (lFlags & qaresult_CANSTEP)
        {
            iError = QA_StepIn (ptSearch->iHandle, iIndex);

            if (iError == 0)
            {
                /* Increment level */
                ptSearch->iLevel ++;

                /* Truncate search history if required */
                if (strlen (sHistory) > ADDRESS_RESULT_WIDTH - 3)
                {
                    strcpy (sHistory + ADDRESS_RESULT_WIDTH - 3, " ..");
                }

                /* Remember new history */
                strcpy (ptSearch->asHistory[ptSearch->iLevel], sHistory);
                

                /* Inform user about the item we've stepped into */
                if (lFlags & qaresult_POSTCODERECODED)
                {
                    QS_DisplayMessage ("Postcode recoded");
                }
                else if (lFlags & qaresult_CROSSBORDERMATCH)
                {
                    QS_DisplayMessage ("Bordering locality");
                }
            }
        }
        else if (lFlags & qaresult_FULLADDRESS)
        {
            ptSearch->iAddress = iIndex;
        }
        else if (lFlags & qaresult_INFORMATION ||
                 lFlags & qaresult_WARNINFORMATION)
        {
            QS_DisplayMessage ("Cannot select item - information only");
        }
        else if (lFlags & qaresult_UNRESOLVABLERANGE)
        {
            QS_DisplayMessage ("Cannot select item - enter value within displayed range");
        }
        else if (lFlags & qaresult_INCOMPLETEADDR)
        {
            QS_DisplayMessage ("Cannot select item - enter premise details");
        }
        else
        {
            QS_DisplayMessage ("Unable to select item");
        }
    }

    return (iError);
}



/**************
* QS_FormatAddress()
*
* Format and display the the selected (leaf-node) address to the user. This
* stage should be modified according to the requirements of the host
* application. Typically, this will result in the selected address being
* written back to the application's data entry screen.
*
* Input:
*   ptSearch   : Search context
*
* Return:
*   0       : Success
*   != 0    : Failure
***************/

int QS_FormatAddress (Search_t *ptSearch)
{
    int     iError  = 0;                    /* QAS Pro error code */
    int     iCount  = 0;                    /* Count of formatted address lines */
    long    lInfo   = 0;                    /* Information about the formatted address */


    /* Format the address */
    if (iError == 0)
    {
        iError = QA_FormatResult (ptSearch->iHandle, ptSearch->iAddress, NULL, &iCount, &lInfo);
    }


    /* Display any warnings to the user */
    if (iError == 0)
    {
        if (lInfo & qaformat_OVERFLOW)
        {
            QS_DisplayMessage ("Warning: There are not enough address lines configured to display the entire address");
        }

        if (lInfo & qaformat_TRUNCATED)
        {
            QS_DisplayMessage ("Warning: Truncation has occurred upon one or more address lines");
        }
    }

    
    /* Display the address */
    if (iError == 0)
    {
        iError = QS_DisplayAddress (ptSearch, iCount);
    }

    return iError;
}


/******************************************************************************
******************* End of file qs_sl.c ***************************************
******************************************************************************/
